/*
 *  test.c
 *  illiConic
 *	John Pacey 2008
 *  
 */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <sys/time.h>    /* for the speedometer */
#include <OpenGL/gl.h>
#include <OpenGL/glu.h>

#if defined(__APPLE__) || defined(__APPLE_CC__)
#include <GLUT/glut.h>
#else
#include <GL/glut.h>
#endif

// flags and constant definitions
/**********************************************************************/

// skel.c flags
#define FLYMODE  (0)
#define TURNMODE (1) 

// tcon.c (illiconic) flags & constants
#define  CONE           2   /* cone is drawn */
#define  DOUBLECONE     4   /* second cone is drawn */
#define  PLANE          8   /* plane is drawn */
#define  VIRG          16   /* virgu/e is drawn */
#define  CURVE          32
#define  WIRE           64
#define  BEAD           128
#define  STARS          256
#define  SPHERE         512
#define  ALTITU        1024
#define  GENERATOR     2048

#define  CNCP           32   /*# of points around cone    - coarse*/
#define  FNCP           64   /*# of points around cone    - fine*/
#define  CNLP           16   /*# of levels on cone        - coarse*/
#define  FNLP           32   /*# of levels on cone        - fine*/
#define  CNLAT          32   /*# of latitudes on spheres  - coarse*/
#define  FNLAT          64   /*# of latitudes on spheres  - fine*/
#define  CNLON          32   /*# of longitudes on spheres - coarse*/
#define  FNLON          64   /*# of longitudes on spheres - fine*/
#define  PS				.5/* planesize */

// flags from tcon.c (currently unused)
#define  SHF            1    /* shiftkey is down */
#define  ALT            2    /* altkey    "   "  */
#define  P2             4    /* PAD2 key is down */
#define  PPL            8    /* PAD+ key is down */
#define  P4             16   /* PAD4 key is down */
#define  P5             32   /* PAD5 key is down */
#define  P6             64   /* PAD6 key is down */
#define  PMI            128  /* PAD- key is down */
#define  P8             256  /* PAD8 key is down */
#define  HK             512  /* HKEY     is down */
#define  KK             1024 /* KKEY     is down */
#define  UPK            2048 /* UPKEY    is down */
#define  DNK            0x1000 /* DNKEY    is down */
#define  RTK            0x2000 /* RTKEY    is down */
#define  LTK            0x4000 /* LTKEY    is down */
#define  JK             0x8000 /* JKEY    is down */

// macros
#define  MAX(x,y)         (((x)<(y))?(y):(x))
#define  MIN(x,y)         (((x)<(y))?(x):(y))
#define  ABS(u)           ((u)<0 ? -(u): (u))
#define  FOR(a,b,c)       for((a)=(b);(a)<(c);(a)++)
#define  DOT(p,q)         ((p)[0]*(q)[0]+(p)[1]*(q)[1]+(p)[2]*(q)[2])
#define  NRM(p)           sqrt(DOT((p),(p)))
#define  DG            M_PI/180
#define  S(u)          sin(u*DG)
#define  C(u)          cos(u*DG)
#define  CLAMP(x,u,v) (x<u? u : (x>v ? v: x))

#define  SGN(x)         ((x)<0?-1:(x)>0?1:0)
#define  SPLIT(x)       (x)>>16, ((x)&0xff00)>>8,(x)&0xff
#define  GETS(www,vvv)  {int inx; for(inx=0;inx<3; inx++) www[inx] = vvv[inx];}

/* function prototypes      *****************************************************/

// main OpenGL loop methods
GLvoid reshaped(int xx, int yy);
GLvoid drawcons(void);

// user-input methods, event handlers
void keyboard(unsigned char key, int x, int y);
void special_keybo(int key, int x, int y);
void chaptrack(int but, int xx, int yy, int shif);
void mousepushed(int but, int stat, int x, int y);
void mousemoved(int x, int y);
void doinput();

// general methods
long atox(const char *str);
void arguments(int argc, char **argv);
void conscone(void);
void conssphere(void);
void autotymer(int reset);
void deFault(void);
float speedometer(void);
void char2wall(float x, float y, float z, char buf[]);
void messages(void);
void idle(void);

// drawing and lighting methods
void drawvert(int th, int ta);
void drawtor(void);
void drawcurve(void);
void drawcircle(float alt, float rad);
void drawcone(void);
void drawwire(void);
void drawsphere(float alt0, float rad, int mode, int colr);
int dolight(int bb, int gg, int rr);
int Hlight(int dog, int cat);
void drawcube(void);
void drawstars(void);
void drawvec(void);
void drawplane(void);
void drawgen(void);
void drawall(void);



/* global variables */

// skel.c variables
float gap, gap0=1.; /* kludge so that arguments() can set a default gap0 */
float lux[3]={1.,2.,3.};             /*world light direction vector     */
float luxx[3];                       /*  object space  direction vector*/
float mysiz,speed, torq, focal, far; /*console navigation variables         */
int win = 2;                   /* 2 full screen, use 0 for demand sized */
unsigned int BUT, XX, YY,SHIF;     /* used in chaptrack gluttery           */ 
int xt,yt;                       /* once was xt,yt,xm,ym for viewportery */
int mode,morph,msg,binoc;            /* pretty global */
int th0, th1, dth, ta0, ta1, dta;    /* torus parameters */

int ii, jj, kk;  float tmp, temp;   /* saves gray hairs later */
float aff[16], starmat[16], mat[16];
float speeds[4]={ 0.007, 0.022, 0.07, 0.22}; 
float nose;   /* to eye distance in console */

// tcon.c (illiconic) variables
int grnd,cube,thick,brt, vuhi, vuhorz, fine,
    curvecolor,conecolor,plancol[4],beadcol[4],wirecolor,gencolor,focsegcolor;

// Hartman movies	
int wri=0, rea=0, all=0, now=0, fcnt=0, frte=1, fnum=0, clinefrte=1;
int odx=0, ody=0, paws=0, opaws=0; /* mouse condition */
int oclaws=0,claws, oflags=0, flags; /* keyflags */

float amb,pwr,
    lmb, bklt, eps, delta, phi, planex, planey, alpha, beta,
    maus,kappa,height,alt0,alt1,psize,coneP[FNCP+1][3],coneN[FNCP+1][3], sphereP[FNLAT+1][FNLON+1][3];

// functions

/**********************************************************************/
void autotymer(int reset){ /* cheap animations */
#define  TYME(cnt,max,act) {static cnt; if(first)cnt=max; else\
                            if(cnt?cnt--:0){ act ; goto Break;}}
  static first = 1;  /* the first time autymer is called */
  if(reset)first=1;  /* or if it is reset to start over  */
  TYME( shrink , 150,th0++;th1--;ta0++;ta1--)
  TYME( pause  , 20,          )
  TYME( grow   , 150,th0--;th1++;ta0--;ta1++)
  TYME( dwell  , 30,            )
  TYME(finish  , 1 , first = 1  )
  first = 0;
  Break:   ;   /* yes Virginia, C has gotos */
}
/**********************************************************************/
void deFault(void){ float tmp; 
th0=5; th1=355;  ta0=5; ta1=355; gap = gap0; 
msg=1; binoc=0; nose=.06; mode=TURNMODE;  
speed=.1; torq=.02; focal = 2.; far =13.; mysiz=.01; morph=0; 
FOR(ii,0,16) starmat[ii]=aff[ii] = (ii/4==ii%4);   /* identities */ 
tmp=NRM(lux); FOR(ii,0,3)lux[ii]/=tmp; amb = .3; pwr = 10. ;
aff[12]=0; aff[13]= 0; aff[14]= -4.2;   /* place where we can see it */
autotymer(1); /* reset autotymer to start at the beginning */

   flags=WIRE+DOUBLECONE+PLANE+CURVE+BEAD+STARS;
   kappa=1.7; height= 1;      /* cone parameters */
   delta = 0.4; phi = 35;     /* plane parameters*/ 
   planex=-0.4; planey=-0.;
   alt0=1; alt1 = -1; eps = .0007; /* beads parameters*/ 
   cube=0; win=1; msg=1; thick=2; fine =0; 
   maus=.005; speed = speeds[2]; 
   mysiz= 0.02 , focal=1.5, far=20; binoc=0;
   vuhi = -17; vuhorz = 1;   /* -34 for console */ 
   grnd = 0; brt = 255; amb= .25; pwr= 16.; nose=.06; bklt=1.;
   plancol[0] = 0x00dd88; plancol[1] = 0x88dd00; 
   plancol[2] = 0x88dd00; plancol[3] = 0x00dd88; 
   curvecolor = 0x0000ff; wirecolor  = 0x11ffdd;
   beadcol[0] = 0xff1111; beadcol[1] = 0xaaaa11;
   beadcol[2] = 0xff8888; beadcol[3] = 0xffff88;
   conecolor  = 0x551188;
   gencolor = 0x33aaff; focsegcolor = 0xaa33ff;
   psize = PS;

   // create cone and sphere
   conscone();
   conssphere();
   
   #ifdef STEREOGRAPHICS
    plancol[0] = 0x000088; plancol[1] = 0x880000;
    plancol[2] = 0x000088; plancol[3] = 0x880000;
    curvecolor = 0x0000ff; wirecolor = 0xdd0011;
    beadcol[0] = 0x1100ff; beadcol[1] = 0x1100aa;
    beadcol[2] = 0x5500ff; beadcol[3] = 0x8800ff;
    conecolor = 0x880055;
   #endif
}

/**********************************************************************/
void drawvert(int th, int ta){  
	float lmb,spec,nn[3], dog, cat;
	nn[0] = C(th)*C(ta); /* unit sphere radius vector */  
	nn[1] = S(th)*C(ta);
	nn[2] =       S(ta);
	lmb = DOT(nn,luxx);lmb =(lmb<0 ? .2 : lmb);  lmb = MAX(amb, lmb); 
	spec = CLAMP((1.1 - pwr+pwr*lmb), 0., 1.);  
	dog = (ta-ta0)/(float)(ta1-ta0); cat = (th-th0)/(float)(th1-th0);
	glColor3f(
		MAX(spec, lmb*dog),                    /* map R^2(dog,cat)->R^3(RGBspace */
		MAX(spec, lmb*(.25 + ABS(cat -.5))),  /* dog cat model of Hartman       */
		MAX(spec, lmb*(1 - cat)));            /* illiLight by Ray Idaszak 1989  */
	glVertex3f(
		C(th) + .5*nn[0],
		S(th) + .5*nn[1],
          .5*nn[2]);
} /* end drawvert */

/**********************************************************************/
void drawtor(void){
int th, ta;
dth = (int)((th1-th0)/24);  /* 24  meridian strips */ 
dta = (int)((ta1-ta0)/24);  /* 24  triangle pairs per strip */
for(th=th0; th < th1; th += dth){
  glBegin(GL_TRIANGLE_STRIP);
  for(ta=ta0; ta < ta1; ta += dta){ drawvert(th,ta); drawvert(th+gap*dth,ta); }
  glEnd();
  } /* end for.theta loop */
}   /* end drawtor */

/**********************************************************************/
void cpack(long hexad)
{
   
   float red, green, blue;
   red   = (hexad & 255)/255.;
   green = ((hexad >> 8) & 255)/255.;
   blue =  ((hexad >> 16)& 255)/255.;
   glColor3f(red,green,blue);
 }

/**********************************************************************/
void drawcurve(void)
{
   int ii,ncp;
   float theta,rad,pt[3];
   ncp=fine?FNCP:CNCP;
   glLineWidth((float)2*thick);
   cpack(curvecolor);
   glBegin(GL_LINE_LOOP);
   FOR(ii,0,ncp)
       {
			theta = M_PI*2*ii/ncp;
			rad = delta/(cos(theta)*sin(phi*DG)+kappa*cos(phi*DG));
			pt[0] = rad * cos(theta);
			pt[1] = rad * sin(theta);
			pt[2] = rad * kappa;
			
			glVertex3fv(pt);
       }
   glEnd();

}

/**********************************************************************/
void drawcircle(float alt,float rad)
{
   int ii,ncp;
   float theta,pt[3];
   ncp=fine?FNCP:CNCP;
   glBegin(GL_LINE_LOOP);
   FOR(ii,0,ncp)
       {
			theta = M_PI*2*ii/ncp;
			pt[0] = rad * cos(theta);
			pt[1] = rad * sin(theta);
			pt[2] = alt; 
			glVertex3fv(pt);
       }
   glEnd();
}

/**********************************************************************/
void conscone(void)
{
	float theta,radius,tmp;
	int ii,inx,ncp;

	ncp=fine?FNCP:CNCP;
	radius = height/kappa; 
/* wrong place for this, put it into drawcone */
/* good luck doing that!  this gets called from many places where drawcone
cannot be called.  Also, drawwire() needs this procedure also, are you going
to put another copy in there?  Besides, it isn't really the wrong place for
this, it is a perfectly good procedure.  It does what the name says.  If you
put it in drawcone, then drawcone does more than draw the cone! */

	FOR(ii,0,ncp+1)
    {
		theta = M_PI*2*ii/ncp;

		coneP[ii][0] = radius * cos(theta);
		coneP[ii][1] = radius * sin(theta);
		coneP[ii][2] = radius * kappa ;

		coneN[ii][0] = coneP[ii][0];
		coneN[ii][1] = coneP[ii][1];
		coneN[ii][2] = -radius/kappa;
		tmp = NRM(coneN[ii]); FOR(inx,0,3) coneN[ii][inx] /= tmp;
    }
}

/**********************************************************************/
void drawcone(void)
{
	int inx, ii,jj, nlp,ncp;
	float tpt[3];
	//float tmp; unused
	
	nlp=fine?FNLP:CNLP;
	ncp=fine?FNCP:CNCP;
	FOR(jj,1,nlp)
    {
		glBegin(GL_TRIANGLE_STRIP);
		FOR(ii,0,ncp+1)
        {
			lmb = DOT(luxx,coneN[ii]);if(lmb<0)lmb= -lmb*bklt;
			cpack(dolight(SPLIT(conecolor)));
			GETS(tpt,coneP[ii]); FOR(inx,0,3) tpt[inx]*=(float)jj/nlp; glVertex3fv(tpt);
			GETS(tpt,coneP[ii]); FOR(inx,0,3) tpt[inx]*=(float)(jj+1)/nlp; glVertex3fv(tpt);        
		}
			glEnd();
     }
}

/**********************************************************************/
void drawwire(void)
{
	int inx, ii,jj, nlp,ncp;
	float tpt[3];
	//float tmp;

	nlp=fine?FNLP:CNLP;
	ncp=fine?FNCP:CNCP;
	glLineWidth((float)thick);
	FOR(jj,1,nlp+1)
    {
		glBegin(GL_LINE_STRIP);
		FOR(ii,0,ncp+1)
        {
			lmb = DOT(luxx,coneN[ii]);if(lmb<0)lmb= -lmb*bklt;
			cpack(dolight(SPLIT(wirecolor)));
			GETS(tpt,coneP[ii]); FOR(inx,0,3) tpt[inx]*=(float)jj/nlp; glVertex3fv(tpt);
        }
		glEnd();
    }
	FOR(ii,0,ncp)
    {
		glBegin(GL_LINE_STRIP);
		FOR(jj,1,nlp+1)
        {
			lmb = DOT(luxx,coneN[ii]);if(lmb<0)lmb= -lmb*bklt;
			cpack(dolight(SPLIT(wirecolor)));
			GETS(tpt,coneP[ii]); FOR(inx,0,3) tpt[inx]*=(float)jj/nlp; glVertex3fv(tpt);
        }
		glEnd();
    }
}

/**********************************************************************/
void conssphere(void)
{
	int ii,jj, nlat,nlon;
	float theta,tau;

	nlat=fine?FNLAT:CNLAT;
	nlon=fine?FNLON:CNLON;
	FOR(jj,0,nlat+1)
    {
		tau = DG*(-80 + jj*160/nlat);
		FOR(ii,0,nlon+1)
        {
			theta = M_PI*2*ii/nlon;
			sphereP[jj][ii][0] = cos(tau) * cos(theta);
			sphereP[jj][ii][1] = cos(tau) * sin(theta);
			sphereP[jj][ii][2] = sin(tau);
        }
    }
}

/**********************************************************************/
void drawsphere(float alt0,float rad, int mode, int colr)
/* mode(wire)=0, mode(skin)=1 */
{
	int ii,jj, nlat,nlon;
	//int inx
	//float theta,tau;

	nlat=fine?FNLAT:CNLAT; 
	nlon=fine?FNLON:CNLON;

	glPushMatrix();
	glTranslatef(0,0,alt0);
	glScalef(rad,rad,rad);

	FOR(jj,0,nlat)
    {
		if(mode==1)glBegin(GL_TRIANGLE_STRIP); else glBegin(GL_LINE_STRIP);
		FOR(ii,0,nlon+1)
        {
			lmb = DOT(luxx,sphereP[jj][ii]);  if(lmb<.0)lmb= -lmb*bklt;
			cpack(dolight(SPLIT(beadcol[colr])));
			glVertex3fv(sphereP[jj][ii]);
			if(mode==0){glEnd(); glBegin(GL_LINE_STRIP);}
			lmb = DOT(luxx,sphereP[jj+1][ii]);if(lmb<.0)lmb= -lmb*bklt;
			cpack(dolight(SPLIT(beadcol[colr])));
			glVertex3fv(sphereP[jj+1][ii]);
        }
		glEnd();
    }
	glPopMatrix();
}

/**********************************************************************/
int dolight(int bb, int gg, int rr)
{
	float spec;

    lmb = MAX(amb,lmb);
    spec=brt*( 1. - pwr + pwr*lmb);
    rr = MAX(lmb*rr, spec);
    gg = MAX(lmb*gg, spec);
    bb = MAX(lmb*bb, spec);
    return( (bb<<16) + (gg<<8)+rr);
}

/**********************************************************************/
void drawvec(void)
{
	float pt[3]={0,0,0};

    glLineWidth((float)2*thick);
    cpack(0xffffaa);
    glPushMatrix();
    glRotatef(phi, 0.0, 1.0, 0.0);
    glBegin(GL_LINE_STRIP); glVertex3fv(pt); pt[2] = delta; glVertex3fv(pt); glEnd();
    glPopMatrix();
}

/**********************************************************************/
void drawplane(void)
{
	float planevt[4][3] = {{-1.,-1., 0.},
                       { 1.,-1., 0.},
                       { 1., 1., 0.},
                       {-1., 1., 0.}};
	int inx;

    glPushMatrix();
    glRotatef(phi, 0.0, 1.0, 0.0);
    glTranslatef(planex,planey,delta);
    glScalef(psize,psize,psize);
    glBegin(GL_POLYGON);
    FOR(inx,0,5){cpack(plancol[inx%4]); glVertex3fv(planevt[inx%4]);} 
    glEnd();
    glPopMatrix();
}

/**********************************************************************/
void drawgen(void) 
{
	float pt[3],sincon,coscon,sinphi,cosphi, rad;
	//float ptC[3], ptH[3];

	/* calculate particulars */
	sinphi = sin(phi*DG); 
	cosphi = cos(phi*DG);
	sincon = 1/sqrt( 1. + kappa*kappa);
	coscon = kappa*sincon;

	glLineWidth((float)2*thick);
	/* top tangency circle */  
	cpack(beadcol[2]);
	drawcircle(alt0*coscon*coscon, alt0*coscon*sincon);
	/* bottom tangency  circle */ 
	cpack(beadcol[3]);
	drawcircle(alt1*coscon*coscon, alt1*coscon*sincon);

	cpack(gencolor);
	glBegin(GL_LINE_STRIP); /* Generator */
	pt[0] = cos(alpha)*height/kappa;
	pt[1] = sin(alpha)*height/kappa;
	pt[2] = height; 
	glVertex3fv(pt);
	pt[0] *= -1;
	pt[1] *= -1;
	pt[2] *= -1;
	glVertex3fv(pt);
	glEnd();

	cpack(focsegcolor);
	glBegin(GL_LINE_STRIP); /* Focus-to-curve segments */
	/*Top focus */
	pt[0] = -alt0*sinphi*sincon;
	pt[1] = 0; 
	pt[2] = alt0*(1. - cosphi*sincon);
	glVertex3fv(pt);
	/* curve point */
	rad = delta/(cos(alpha)*sinphi + kappa*cosphi);
	pt[0] = rad * cos(alpha);
	pt[1] = rad * sin(alpha);
	pt[2] = rad * kappa;
	glVertex3fv(pt);
	/* bot focus */
	pt[0] = alt1*sinphi*sincon;
	pt[1] = 0;
	pt[2] = alt1*(1. + cosphi*sincon);
	glVertex3fv(pt);
	glEnd();

	return;
}

/**********************************************************************/
void drawcube(void)
{ int ii,jj ; /* Steve Kommrusch, 1984, uses Gray code */ 
    static int firstime=1; static float vq[8][3];  
    static int fr[]= {0,1,5,1,3,7,3,2,6,2,0,4,5,7,6,4}; 
    if(firstime){ FOR(ii,0,8)FOR(jj,0,3)vq[ii][jj]=(ii&(1<<jj))?PS:-PS;
            firstime=0;}
    //lRGBrange(99,0,99,255,0,255,znear,zfar);  ******** need to translate this and depthcue calls
    glLineWidth((float)thick); 
    //depthcue(1);  *******
    glBegin(GL_LINE_STRIP); 
	FOR(ii,0,16) glVertex3fv(vq[fr[ii]]); 
	glEnd();
    //depthcue(0); *********
}

/**********************************************************************/
/* Idaszak-Hartman paintor*/
int Hlight(int dog,int cat)
{ return(dolight(dog,abs(cat-128)+64,brt-cat)); }
 
/**********************************************************************/
void drawall(void)
{
	//drawtor(); drawcube();
	
	float sincone;  /* kappa=cot_cone_angle */
	sincone = 1/sqrt(1. + kappa*kappa);

	if(flags&ALTITU||flags&GENERATOR)
	{
		alt0 =  delta/(cos(phi*DG) - sincone); 
		alt1 =  delta/(cos(phi*DG) + sincone); 
	}
	
	if(cube)
		drawcube();
   	
	if(flags&CONE) 
	{
		drawcone();
		if(flags&DOUBLECONE)
		{
			glScalef(-1,-1,-1);
			drawcone();
			glScalef(-1,-1,-1);
		}
	}
	
	if(flags&PLANE) 
		drawplane();
	
	if(flags&VIRG)   
		drawvec();
	
	if(flags&BEAD)
	{
       drawsphere(alt0, eps*SGN(alt0) + alt0*sincone, 1, 0);
       drawsphere(alt1, eps*SGN(alt1) + alt1*sincone, 1, 1);
	}

	if(flags&SPHERE)
	{
		drawsphere(alt0, alt0*sincone, 0, 2);
		drawsphere(alt1, alt1*sincone, 0, 3);
	}
	
	if(flags&CURVE) 
		drawcurve();
	
	if(flags&GENERATOR) 
		drawgen();
	
	if(flags&WIRE)
	{
		drawwire();
		if(flags&DOUBLECONE)
		{
			glScalef(-1,-1,-1);
			drawwire();
			glScalef(-1,-1,-1);
		}
	}
}

/**********************************************************************/
void drawstars(void){ 
     static float star[1000][3]; static int virgin=1;
 if(virgin){  /* first time  set up the stars */
     FOR(ii,0,1000)FOR(jj,0,3)star[ii][jj]=random()/(float)0x40000000-1.;
     virgin=0; /* never again */
  }
  glMatrixMode(GL_MODELVIEW);
  glPushMatrix();   /* insurance of superstition */
  glMultMatrixf(starmat);
  glColor3f(1.0,1.0,0.0);    
  glBegin(GL_POINTS);
    FOR(ii,0,1000)glVertex3fv(star[ii]);
  glEnd();
  glPopMatrix();
  glClear(GL_DEPTH_BUFFER_BIT); /* so the stars are behind everything */
}

/**********************************************************************/
long atox(const char *str) /* Loren Evans ? or Chris Hartman */
{
    long num = 0;
    int radix = 10;
    int val = 0;
    if ( str[0] == '0' && tolower(str[1]) == 'x' )  {
        radix = 16;
        str += 2;
    }
    while ( *str )  {
        if ( !isxdigit(*str) )  return 0;
        switch( tolower(*str) )  {
            case 'a':  val = 10; break;
            case 'b':  val = 11; break;
            case 'c':  val = 12; break;
            case 'd':  val = 13; break;
            case 'e':  val = 14; break;
            case 'f':  val = 15; break;
            default :  val = ( *str - '0' );
        }
        num = num*radix + val;
        str++;
    }
    return num;
}

/**********************************************************************/
void arguments(int argc,char **argv){
  extern char *optarg;
  extern int optind; 
  int chi;   
  /* w: needs ONE number after -w, c<nothing> means NO number follows*/ 
  while ((chi = getopt(argc,argv,"w:d:g:t:c:v:f:b:p:u:m:s")) != -1) 
  switch(chi)  {
                case 'w': win=atoi(optarg); break; 
                case 'g': gap0=atof(optarg); break;
				case 't': thick=atoi(optarg); break;
				case 'c': conecolor=atox(optarg); break;
				case 'v': curvecolor=atox(optarg); break;
				case 'f': wirecolor=atox(optarg); break;
				case 'b': break;
				case 'p': break;
				case 'u': break;
				case 'm': frte=atoi(optarg);  break;
				case 's': psize=atof(optarg); break;
                }
  if (optind!=argc) fprintf(stderr,"%s: Incorrect usage\n",argv[0]);
}

/*************************************************************************/

/**********************************************************************/
void keyboard(unsigned char key, int x, int y){
#define  IF(K)				if(key==K)
#define  PRESS(K,A,b)		IF(K){b;}IF((K-32)){A;}  
/*was backwards in previous versions */
#define  BITOGGLE(K,f,b)	IF(K){f^=b;}
#define  TOGGLE(K,flg)		IF(K){(flg) = 1-(flg); }
#define  CYCLE(K,f,m)		PRESS((K), (f)=(((f)+(m)-1)%(m)),(f)=(++(f)%(m)))
#define  SLIDI(K,f,m,M)		PRESS(K,(--f<m?m:f), (++f>M?M:f))
#define  SLIDF(K,f,m,M,d)	PRESS(K,((f -= d)<m?m:f), ((f += d)>M?M:f))
#define  PRESS2(K,f,m)      PRESS(K,f,m) conscone();
#define  IFSHIFT(Y,N)		if(glutGetModifiers()==GLUT_ACTIVE_SHIFT){ Y;} else {N;}
#define  IFALT(Y,N)			if(glutGetModifiers()==GLUT_ACTIVE_ALT){Y;} else {N;}
#define  VERNIER(K,var,amt) {float v=1; IFALT(v=10,) PRESS(K,var-=amt/v/frte, var+= amt/v/frte)}
#define  VERNIE2(K,var,amt) {float t=1; IFALT(t/=10,) PRESS(K, var-=t*amt,var+=t*amt)}

/* Only ASCII characters can be processes by this GLUT callback function */ 
   IF(27) { exit(0); }							 /* ESC exit             */
   PRESS('9', vuhi--,vuhi++)					 /* for CAVE  */
   PRESS('0', vuhorz--, vuhorz++)				 /* for CAVE */
   TOGGLE('m',msg);                              /* writing on/off        */
   TOGGLE('v',binoc);                            /* cross-eyed STEREO     */
   TOGGLE('*',grnd);							 /* white for the printer */
   CYCLE('q',cube,2);							 /* stick cube appears    */
   PRESS('z', deFault(), deFault());             /* zap changes(reset)    */
   PRESS('r',far*=1.01,far/=1.01);				 /* beware of this        */
   PRESS('n', nose -= .001 , nose += .001 );     /* for binoculars        */
   PRESS('e',eps-=.01,eps+=.01);				 /* an epsilon            */
   PRESS('t',thick=(thick==1?1:thick-1),thick++) /* wireframe thickness adjuster */
   PRESS('f',{fine=0;conscone();conssphere();}
		,{fine=1;conscone();conssphere();})		 /* course/fine mesh  */	
		
   //PRESS('g',gap /= .9, gap *= .9);              /* gap parameter        */
   
/* the following inputs are movie recorded */
   PRESS('i', mysiz /= 1.1, mysiz *= 1.1)        /* rescale the world       */
   PRESS('o', focal *= 1.1 , focal /= 1.1)       /* telephoto i.e. zoom out */
   PRESS('1',speed /= 1.02, speed *= 1.02);      /* flying speed         */
   PRESS('3',torq /= 1.02, torq *= 1.02);        /* turning speed        */
   TOGGLE('u',morph);                            /* autotymer on/off        */
   
   // clear flag variables
   claws=0, paws=0;
   
   BITOGGLE('`', flags,STARS)					/* turn stars on or off */
   BITOGGLE('c',flags,CONE)                     /* coneskin drawn   */    
   BITOGGLE('w',flags,WIRE)                     /* conewire drawn   */ 
   BITOGGLE('b',flags,BEAD)                     /* sphere rendered  */ 
   BITOGGLE('s',flags,SPHERE)                   /* sphere drawn     */ 
   BITOGGLE('a',flags,ALTITU)                   /* free altitudes   */ 
   BITOGGLE('p',flags,PLANE)                    /* plane drawn      */ 
   BITOGGLE('d',flags,DOUBLECONE)               /* two cones drawn  */ 
   BITOGGLE('g',flags,GENERATOR)                /* generator  drawn */ 
   BITOGGLE('/',flags,VIRG)						/* normal  drawn    */
   BITOGGLE('x',flags,CURVE)			        /* curve  drawn     */
   
   VERNIER('k', kappa, .05)						/*    kappa         */
   VERNIER('y', psize, -.05)
   PRESS('h', height-=.05, height+=.05)			/*    height        */
   conscone();  //recompute cone
   CYCLE(' ', mode,TURNMODE+1);                 /*  fly/turn modes  */
   
   // plane movement relative to initial config with blue sphere on top 
   VERNIER('2', planex, .05)					/*  Move plane South- ("down") */
   VERNIER('8', planex, -.05)					/*  Move plane North- ("up") */
   VERNIER('6', planey, .05)					/*  Move plane East- ("right") */
   VERNIER('4', planey, -.05)					/*  Move plane West- ("left") */
   
   VERNIER('+', delta, .05)					   
   VERNIER('-', delta, -.05)					
   VERNIER('5', phi, 5.)						
   VERNIER('j', beta, 5.)					
}
/**********************************************************************/
void special_keybo(int key, int x, int y){
/*non-ASCII keypresses go here, if you're lucky enough to know their names */

 //fprintf(stderr," non-ASCII character was pressed.\n");
 //fprintf(stderr," use special_keybo() to process it\n");
 
 VERNIE2(GLUT_KEY_F1, amb, .01)					/* ambient floor		*/
 PRESS(GLUT_KEY_F2, pwr= 1. , pwr *= 2)         /* specular ramp		*/ 
 VERNIER(GLUT_KEY_RIGHT, alpha, .1)				/*						*/
 VERNIER(GLUT_KEY_LEFT, alpha, -.1)				/*						*/
 VERNIER(GLUT_KEY_UP, alt0, .05)				/* blue sphere origin	*/	
 VERNIER(GLUT_KEY_DOWN, alt1, .05)				/* cyan sphere origin	*/
 
 
}

/**********************************************************************/

/**********************************************************************/
float speedometer(void){
double dbl; static double rate; static int ii=0;
static struct timezone notused; static struct timeval now, then;
   if(++ii % 8 == 0){  /* 8 times around measure time */
      gettimeofday(&now, &notused); /* elapsed time */
      dbl =  (double)(now.tv_sec - then.tv_sec)
         +(double)(now.tv_usec - then.tv_usec)/1000000;
      then = now;  rate = 8/dbl;
      }
   return((float)rate);
}
/**********************************************************************/
void char2wall(float x,float y,float z, char buf[]){
     char *p; glRasterPos3f(x,y,z);
     for(p = buf;*p;p++) glutBitmapCharacter(GLUT_BITMAP_9_BY_15,*p);
}
/**********************************************************************/
void messages(void){     
  char buf[256]; /* console messages are done differently from cave */
	#define  LABEL1(x,y,W,u) {sprintf(buf,(W),(u));char2wall(x,y,0.,buf);}
	#define  LABEL2(x,y,W,u,v) {sprintf(buf,(W),(u),(v));char2wall(x,y,0.,buf);}
  glMatrixMode(GL_PROJECTION); glPushMatrix(); glLoadIdentity();
  gluOrtho2D(0,3000,0,3000);
  glMatrixMode(GL_MODELVIEW); glPushMatrix(); glLoadIdentity();
      /*bull's eye*/
      if(mode==TURNMODE) glColor3f(0x22/255.,0x88/255.,0xdd/255.);
        else      glColor3f(.8,.8,.8);
      LABEL1(1500,1500,"%s","o");
      /* writings */
      if(mode==TURNMODE) glColor3f(1.,0.,1.);
        else      glColor3f(1.,1.,0.);
      LABEL1(80,80,"%4.1f fps",speedometer());
      LABEL1(80,2840,"(ESC)ape   (Z)ap","");
	  LABEL1(900,2840,"(BAR)fly  (V)iew  (L)ine  (`)stars  (Q)ube" ,"");
      LABEL1(2000,2840,"a(U)totymer co(M)mands","");
	  LABEL1(10,10,"illiconic \
   by Pacey, Francis, Bourd, Hartman & Chappell, U Illinois, 1995..1997,2008 %s","");
	  LABEL1(1900,200," the shifted key inverts the action, usually","");
      LABEL1(80,2770,"(C)one  (W)ire  (D)ouble","");
      LABEL2(80,2700,"(K)appa %5.3f (H)eight %5.3f", kappa, height);
      LABEL1(80,2630,"(P)lane size(Y) move on PAD keys:","");
      LABEL1(80,2560,"norma(/) phi(5)ve %5.3f",phi);
      LABEL1(80,2490,"(8)(2)(6)(4) NSEW","");
	  LABEL1(80,2420,"(B)ead(S) (A)ltitudes:","");
	  LABEL1(80,2350,"Blue(UP) %5.3f",alt0);
	  LABEL1(80,2280,"Cyan(DN) %5.3f",alt1);
	  LABEL1(80,2210,"(E)psilon  %g",eps);
	 
	  LABEL2(2000,2770,"(F1)amb %g (F2)spec %g", amb, pwr);
      LABEL1(2000,2700,"(N)ose %g",nose);
      LABEL1(2000,2630,"(1) speed  %g",speed);
      LABEL1(2000,2560,"near clipper %g", mysiz*focal);
      LABEL1(2000,2490,"f(O)cal factor %g",focal);
	  LABEL1(2000,2420,"my s(I)ze %g",mysiz);
	  LABEL1(2000,2350,"far clippe(R)= %g",far);
	  LABEL1(2000,2280,"(0)vuhorz %d",vuhorz);
	  LABEL1(2000,2210,"(9)vuhi  %d",vuhi);

	  
    glPopMatrix();
	glMatrixMode(GL_PROJECTION); glPopMatrix();
} 
/**********************************************************************/
void chaptrack(int but,int xx,int yy,int shif){  
   long dx,dy; 
   dx = xx -.5*xt; dx = abs(dx)>5?dx:0;
   dy = yy -.5*yt; dy = abs(dy)>5?dy:0;

   glMatrixMode(GL_MODELVIEW); glPushMatrix(); glLoadIdentity();

   if(mode==TURNMODE) glTranslatef(aff[12],aff[13],aff[14]);
   glRotatef(dx*torq,0.,1.,0.); glRotatef(dy*torq,1.,0.,0.);
   if(but&(1<<GLUT_RIGHT_BUTTON ))glRotatef(shif?-10:-1,0.,0.,1.);
   if(but&(1<<GLUT_LEFT_BUTTON  ))glRotatef(shif?10:1,0.,0.,1.);
   if(mode==FLYMODE){
      glPushMatrix();
      glMultMatrixf(starmat);
      glGetFloatv(GL_MODELVIEW_MATRIX,starmat);
      glPopMatrix();
   }
   if(but&(1<<GLUT_MIDDLE_BUTTON)) glTranslatef(0.,0.,shif ? -speed : speed);
   if(mode==TURNMODE) glTranslatef(-aff[12],-aff[13],-aff[14]);
   glMultMatrixf(aff); 
   glGetFloatv(GL_MODELVIEW_MATRIX,aff);
   FOR(ii,0,3){luxx[ii]=0; FOR(jj,0,3)luxx[ii] +=aff[ii*4+jj]*lux[jj];}
   glPopMatrix();
}
/**********************************************************************/
GLvoid reshaped(int xx, int yy){xt=xx ; yt=yy;}   /* origin of moved window */

/**********************************************************************/
GLvoid drawcons(void)
{  
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  if(binoc) glViewport(0,yt/4,xt/2,yt/2);
  glMatrixMode(GL_PROJECTION); glLoadIdentity();
  glFrustum(-mysiz*xt/yt,mysiz*xt/yt,-mysiz,mysiz,mysiz*focal,far); 
  glMatrixMode(GL_MODELVIEW); glLoadIdentity();
  if(flags&STARS)
	drawstars();
  glTranslatef(-binoc*nose,0.0,0.0);
  glMultMatrixf(aff);
  glRotatef(beta,0.0,0.0,1.0);
  drawall();
  if(binoc){
    glViewport(xt/2,yt/4,xt/2,yt/2); 
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
	if(flags&STARS)
		drawstars();
    glTranslatef(binoc*nose,0.0,0.0);
    glMultMatrixf(aff);
	glRotatef(beta,0.0,0.0,1.0);
    drawall();
    }
  glViewport(0,0,xt,yt);
  if(msg) messages();
  glutSwapBuffers();
}

/**********************************************************************/
void idle(void){ /*do this when nothing else is happening*/
   if(morph) autotymer(0);  /* advance autotymer */ 
   glutPostRedisplay();  /*redraw the window*/
   chaptrack(BUT,XX,YY,SHIF);
}

/**********************************************************************/
void mousepushed(int but,int stat,int x,int y){
  if(stat==GLUT_DOWN) BUT |= (1<<but);
  else BUT &= (-1 ^ (1<<but));  
  XX=x; YY=y; SHIF=(glutGetModifiers()==GLUT_ACTIVE_SHIFT)?1:0;
  if (but==GLUT_LEFT_BUTTON)
	paws|=1;
  if(but==GLUT_RIGHT_BUTTON)
	paws|=2;
  if(but==GLUT_MIDDLE_BUTTON)
	paws|=4;
}

/**********************************************************************/
void mousemoved(int x,int y){  XX=x; YY=y; }
/**********************************************************************/
int main(int argc, char **argv){  
   arguments(argc,argv); 
   deFault(); 
 
       glutInit(&argc, argv);
       glutInitDisplayMode(GLUT_DOUBLE|GLUT_DEPTH);
       switch(win){ 
           case 0: break;  /* manage your own window */
           case 1: glutInitWindowSize(640, 480);
                   glutInitWindowPosition(100,100); break;
           case 2: glutInitWindowPosition(0,0); break;
         }
       glutCreateWindow("<* illiconix in C/OpenGL/GLUT *>");
       if(win==2) glutFullScreen();
       glEnable(GL_DEPTH_TEST);
       glutDisplayFunc(drawcons);       
       glutKeyboardFunc(keyboard);
       glutSpecialFunc(special_keybo);
       glutMouseFunc(mousepushed);
       glutMotionFunc(mousemoved);       
       glutPassiveMotionFunc(mousemoved); 
       glutReshapeFunc(reshaped);
       glutIdleFunc(idle);             
       glutMainLoop();
	   return 0;
}
 
